/*****************************************************************************
*
* Atmel Corporation
*
* File              : USI_TWI_Slave.c
* Compiler          : IAR EWAAVR 2.28a/3.10a
* Revision          : $Revision: 1.7 $
* Date              : $Date: 25. mai 2004 14:51:12 $
* Updated by        : $Author: ltwa $
*
* Support mail      : avr@atmel.com
*
* Supported devices : All device with USI module can be used.
*                     The example is written for the ATmega169, ATtiny26 & 
*                     ATtiny2313
*
* AppNote           : AVR312 - Using the USI module as a TWI slave
*
* Description       : Functions for USI_TWI_receiver and USI_TWI_transmitter.
*
* 
****************************************************************************/

#include <iom168v.h>
#include <macros.h>
#include "USI_TWI_Slave.h"

//#pragma interrupt_handler  USI_Start_Condition_ISR:iv_USI_START
//#pragma interrupt_handler  USI_Counter_Overflow_ISR:iv_USI_OVF

//********** Static Variables **********//

static unsigned char TWI_slaveAddress;

static unsigned char TWI_RxBuf[TWI_RX_BUFFER_SIZE];
static unsigned char TWI_TxBuf[TWI_TX_BUFFER_SIZE];

// USI_TWI_S_statusRegister is defined in USI_TWI_Slave_config.h
struct USI_TWI_S_statusReg USI_TWI_statusReg = {0};

// Status byte holding flags.  
struct USI_TWI_state                          
{
    unsigned addressMode:1;
    // masterReadDataMode = 0 is equal to "writeDataMode" = 1
    unsigned masterReadDataMode:1;   
    unsigned unused:6;
} USI_TWI_state = {0};

extern unsigned char isrtrigger,isrcnt;


//********** USI_TWI functions **********//

/*----------------------------------------------------------
  Initialise USI for TWI Slave mode.
----------------------------------------------------------*/

void USI_TWI_Slave_Initialise(unsigned char TWI_ownAddress)
{
    TWI_slaveAddress = TWI_ownAddress;

    PORT_USI |=  (1<<PORT_USI_SCL);    // Set SCL high
    PORT_USI |=  (1<<PORT_USI_SDA);    // Set SDA high
    DDR_USI  |=  (1<<PORT_USI_SCL);    // Set SCL as output    
//    DDR_USI  &= ~(1<<PORT_USI_SCL);    // Set SCL as input   
    DDR_USI  &= ~(1<<PORT_USI_SDA);    // Set SDA as input
	
	// Enable Start Condition Interrupt. Disable Overflow Interrupt.
	// Set USI in Two-wire mode. No USI Counter overflow prior 
	// to first Start Condition (potentail failure)
	// Shift Register Clock Source = External, positive edge    
    USICR    =  (1<<USISIE)|(0<<USIOIE)| 
                (1<<USIWM1)|(0<<USIWM0)|                      
                (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|  
                (0<<USITC);
	// Clear all flags and reset overflow counter		  
    USISR    = 0xF0;         
}

/*----------------------------------------------------------
 Functions for filling and tapping the USI_TWI buffers.
----------------------------------------------------------*/
unsigned char USI_TWI_Start_Transceiver_With_Data(unsigned char *msg,
		 	  									  unsigned char msgSize)
{
    unsigned char i;

#ifdef BUFFER_CONTROL
    if(msgSize > TWI_TX_BUFFER_SIZE)
        return (FALSE);
#endif

    if(!USI_TWI_statusReg.dataInTxBuf)  {
        for(i=0; i<msgSize; i++)
             TWI_TxBuf[i] = msg[i];
        USI_TWI_statusReg.dataInTxBuf = TRUE;
        return (TRUE);
        }
    else
        return (FALSE);
}


unsigned char USI_TWI_Get_Data_From_Transceiver(unsigned char *msg,
                                                unsigned char msgSize)
{
    unsigned char i;

#ifdef BUFFER_CONTROL
    if(msgSize > TWI_RX_BUFFER_SIZE)
        return (FALSE);
#endif

    if(USI_TWI_statusReg.dataInRxBuf)  {
        for(i=0; i<msgSize; i++)
             msg[i] = TWI_RxBuf[i];
        USI_TWI_statusReg.dataInRxBuf = FALSE;
        return (TRUE);
        }
    else
        return (FALSE);
}

// ********** Interrupt Handlers ********** //

/*----------------------------------------------------------
 Detects the USI_TWI Start Condition and intialises the USI 
 for reception of the "TWI Address" packet.
----------------------------------------------------------*/
void USI_Start_Condition_ISR(void)
{ 
    // Set default starting conditions for new TWI package
	// Expect TWI address on bus
    USI_TWI_state.addressMode = TRUE;                               
    DDR_USI  &= ~(1<<PORT_USI_SDA);  // Set SDA to input
	
	// Added by SS   
	while(PINB&(1<<PB2))  ;  // wait for SCL line to go low

    // Enable TWI Counter overflow and interrupts
	// Enable Overflow and Start Condition Interrupt. 
	// (Keep StartCondInt to detect RESTART)
	// Set USI in Two-wire mode.
	// Shift Register Clock Source = External, positive edge
    USICR   =   (1<<USISIE)|(1<<USIOIE)|                            
                (1<<USIWM1)|(1<<USIWM0)|                            
                (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|                
                (0<<USITC);
	// Clear flags
	// Set USI to sample 8 bits i.e. count 16 external pin toggles
	// In theory setting USISIF should clear the SCL low condition
	// but it doesn't
    USISR   =   (1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|      
                (0x0<<USICNT0); 
}


/*----------------------------------------------------------
 Handles all the comunication. Is disabled only when waiting 
 for new Start Condition.
----------------------------------------------------------*/
void USI_Counter_Overflow_ISR(void)
{
    static volatile unsigned char TWI_bufPtr;

    // Inspect Address and R/W bit and set flags accordingly; ACK knowledge on 
    // address if able to comply to the type of operation requested by the 
	// master (read or write) - depends on whether data is available for 
	// transmission - if there is room for received data
    if(USI_TWI_state.addressMode) {  // If expecting address packet  
		
        // If the address is a General Call or equal to mine, then start 
		// requested operation (R/W).
        if((USIDR==0) || ((USIDR>>1)==TWI_slaveAddress)) {
		    // Initialize Buffer to receive/transmit data
            TWI_bufPtr = 0;                                            
            if(USIDR == 0)  USI_TWI_statusReg.genAddressCall = TRUE;
            else  USI_TWI_statusReg.genAddressCall = FALSE;
            
			// Use R/W DataMode from now on.
            USI_TWI_state.addressMode = FALSE;      
			// USIDR_LSB = R/W Mode. FALSE => masterWriteDataMode                    
            USI_TWI_state.masterReadDataMode = (USIDR & 0x01); 
			// Store Read or Write status in first byte of rx buffer 
			TWI_RxBuf[TWI_bufPtr++] = (USIDR & 0x01);
			  
            USIDR = 0;    // Prepare ACK                                              
            
            if(USI_TWI_state.masterReadDataMode) {
			    // If the TxBuf is empty when Master requests data - 
			    // To NACK on address just jump straight to
			    // Start Condition detection mode
                if(!USI_TWI_statusReg.dataInTxBuf) {
				    // Enable Start Condition Interrupt. Disable Overflow 
					// Interrupt. Set USI in Two-wire mode. No USI Counter 
					// overflow hold. Shift Register Clock Source = External, 
					// positive edge
                    USICR    =  (1<<USISIE)|(0<<USIOIE)|                
                                (1<<USIWM1)|(0<<USIWM0)|                
                                (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|    
                                (0<<USITC);
					// Clear all flags but Start Condition
                    USISR   =   (0<<USISIF)|                            
                                (1<<USIOIF)|
                                (1<<USIPF)|
                                (1<<USIDC)|      
                                (0x0<<USICNT0);
                    return;
                	}
            	}
			// Must be master write command
            // If RxBuf is full when Master is sending NACK on address 
			// and jump straight to Start Condition detection mode
            else if(USI_TWI_statusReg.dataInRxBuf) {
				// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				// Set USI in Two-wire mode. No USI Counter overflow hold.
				// Shift Register Clock Source = External, positive edge
                USICR    =  (1<<USISIE)|(0<<USIOIE)|                
                            (1<<USIWM1)|(0<<USIWM0)|                
                            (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|    
                            (0<<USITC);
				// Clear all flags but Start Condition
                USISR   =   (0<<USISIF)|                            
                            (1<<USIOIF)|
                            (1<<USIPF)|
                            (1<<USIDC)|      
                            (0x0<<USICNT0);
                return;
            	}
			
			//  Otherwise address matches and buffers are ready
			//  to send/receive data
			USIDR = 0;
            DDR_USI |=  (1<<PORT_USI_SDA);    // Set SDA as output 
			// Clear all flags, except Start Cond 
			// Set USI to shift out 1 bit                          
            USISR   =   (0<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|  
                        (0xE<<USICNT0);       
			// Wait until ACK sent. (Pass if Start Cond.)
            while(!(USISR & ( (1<<USISIF)|(1<<USIOIF))));           
            DDR_USI &=  ~(1<<PORT_USI_SDA);   // Set SDA as input
            }

        // If received address did not match own address - ignore package. 
		// Wait for new Start Condition
        else  {      
			// Set USI in TWI Start Condition Mode
			// Enable Start Condition Interrupt. Disable Overflow Interrupt.
			// Set USI in Two-wire mode. No USI Counter overflow hold. 
			// Shift Register Clock Source = External, positive edge                                                       
            USICR    =  (1<<USISIE)|(0<<USIOIE)|            
                        (1<<USIWM1)|(0<<USIWM0)|                       
                        (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|           
                        (0<<USITC);
			// Clear all flags but Start Condition
            USISR   =   (0<<USISIF)|                                   
                        (1<<USIOIF)|
                        (1<<USIPF)|
                        (1<<USIDC)|      
                        (0x0<<USICNT0);
            return;
        	}
    	}

    // Address already accepted
	// Decide if this is Read or Write command and (N)ACK
    else  {
        // One or more bytes transmitted by slave (Master Read)
		// Read NACK if transmission completed or ACK to continue 
        if(USI_TWI_state.masterReadDataMode)   {
            DDR_USI &=  ~(1<<PORT_USI_SDA);    // Set SDA as input
            USIDR = 0;
			// Clear all flags, except Start Cond
			// Set USI to sample 1 bit
            USISR = (0<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|  
                    (0xE<<USICNT0);
			// Wait until (N)ACK received. (Pass if Start Cond.)                                 
            while(!(USISR & ((1<<USISIF)|(1<<USIOIF)))); 
			// If NACK (USIDR_LSB=1) re-init the Start Cond detector          
            if(USIDR)  {
                USI_TWI_statusReg.dataInTxBuf = FALSE;
				// Enable Start Condition Interrupt. Disable Overflow Interrupt
				// Set USI in Two-wire mode. No USI Counter overflow hold
				// Shift Register Clock Source = External, positive edge
                USICR    =  (1<<USISIE)|(0<<USIOIE)|                
                            (1<<USIWM1)|(0<<USIWM0)|                
                            (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|    
                            (0<<USITC);
				// Clear all flags but Start Condition
                USISR   =   (0<<USISIF)|                            
                            (1<<USIOIF)|
                            (1<<USIPF)|
                            (1<<USIDC)|      
                            (0x0<<USICNT0);
                return;
            	}

#ifdef BUFFER_CONTROL
            // If the Master (has sent ACK and) is requesting more 
			// information than the size of TxBuf, then set OverflowFlag 
			// and terminate
            else  {
				// But if the transmit buffer is full, then set Overflow flag
                if(TWI_bufPtr == TWI_TX_BUFFER_SIZE)  {
                    USI_TWI_statusReg.TxBufOverflow = TRUE;
                    // Set USI in TWI Start Condition Mode
					// Enable Start Condition Interrupt. Disable Overflow 
					// Interrupt. Set USI in Two-wire mode. No USI Counter 
					// overflow hold.  Shift Register Clock Source = External, 
					// positive edge
                    USICR    =  (1<<USISIE)|(0<<USIOIE)|                
                                (1<<USIWM1)|(0<<USIWM0)|                
                                (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|    
                                (0<<USITC);
					// Clear all flags but Start Condition
                    USISR   =   (0<<USISIF)|                            
                                (1<<USIOIF)|
                                (1<<USIPF)|
                                (1<<USIDC)|      
                                (0x0<<USICNT0);
                    return;
                	}
            	}
#endif
             }

		// Must be "masterWriteDataMode" - Slave receive. Store the 
        // sampled data and send an ACK to the Master.
        else  {
            TWI_RxBuf[TWI_bufPtr++] = USIDR;
            USIDR = 0;      // Send ACK

#ifdef BUFFER_CONTROL 
	   		// If the receive buffer is full, then send NACK.
            if(TWI_bufPtr == TWI_RX_BUFFER_SIZE)  {
                USI_TWI_statusReg.RxBufOverflow = TRUE;
                USIDR = 0x80;    // send NACK
                }
#endif
            DDR_USI |=  (1<<PORT_USI_SDA);   // Set SDA to output
			// Clear all flags, except Start Cond
			// Set USI to shift out 1 bit
            USISR   =   (0<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|  
                        (0xE<<USICNT0); 
			// Wait until (N)ACK sent. (Pass if Start Cond.)                                
            while(!(USISR & ( (1<<USISIF)|(1<<USIOIF))));           
            DDR_USI &=  ~(1<<PORT_USI_SDA);    // Set SDA as intput
            }
        }

    // If The TWI master has initiated a read cycle (slave transmit) 
	// copy data from buffer to USIDR
    if(USI_TWI_state.masterReadDataMode)  {
	    // Put data in USI Data Register.
        USIDR   =   TWI_TxBuf[TWI_bufPtr++];                            
        DDR_USI |=  (1<<PORT_USI_SDA);   // Set SDA as output
		// Clear all flags, except Start Cond
		// Set USI to shift out 8 bits
        USISR   =   (0<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|      
                    (0x0<<USICNT0);                                     
        }

    // "masterWriteDataMode". I.e. prepare to sample a data byte. 
	// (Or stop condition.)
    else  {
//        DDR_USI &=  ~(1<<PORT_USI_SDA);   // Set SDA as input

        // Clear all flags, except Start Cond
		// Set USI to sample 1 bit
        USISR   =   (0<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|      
                    (0xE<<USICNT0);  
		// Wait until 1 bit sampled or STOP condition. (Pass if Start Cond.) 
        while(!(USISR & ((1<<USISIF)|(1<<USIOIF)|(1<<USIPF))));    
              
		// If STOP Cond then set USI in START Cond Mode       
        if(USISR & (1<<USIPF))  {
		    // Transmission complete i.e. valid data in RxBuf
            USI_TWI_statusReg.dataInRxBuf = TRUE;
			// Load number of received bytes into first buffer offset by 4
			TWI_RxBuf[0] = TWI_RxBuf[0] | ((TWI_bufPtr-1)<<4);
			                       
            // Set USI in TWI Start Condition Mode
			// Enable Start Condition Interrupt. Disable Overflow Interrupt.
			// Set USI in Two-wire mode. No USI Counter overflow hold.
			// Shift Register Clock Source = External, positive edge
            USICR    =  (1<<USISIE)|(0<<USIOIE)|                        
                        (1<<USIWM1)|(0<<USIWM0)|                        
                        (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|            
                        (0<<USITC);
			// Clear all flags but Start Condition
            USISR   =   (0<<USISIF)|                                    
                        (1<<USIOIF)|
                        (1<<USIPF)|
                        (1<<USIDC)|      
                        (0x0<<USICNT0);
            return;
            }
        else {
		    // Clear all flags, except Start Cond
			// Set USI to sample last 7 bits
            USISR   =   (0<<USISIF)|(1<<USIOIF)|(0<<USIPF)|(1<<USIDC)|  
                        (0x2<<USICNT0);                                 
            }
        }
}
